@@ -1,5 +1,3 @@
-from __future__ import absolute_import, division, print_function
-
 import asyncio
 import io
 import logging
@@ -107,6 +105,10 @@ class HTTPFileSystem(AsyncFileSystem):
         request_options.pop("skip_instance_cache", None)
         self.kwargs = request_options
 
+    @property
+    def fsid(self):
+        return "http"
+
     def encode_url(self, url):
         return yarl.URL(url, encoded=self.encoded)
 
@@ -163,7 +165,7 @@ class HTTPFileSystem(AsyncFileSystem):
                 l = l[1]
             if l.startswith("/") and len(l) > 1:
                 # absolute URL on this server
-                l = parts.scheme + "://" + parts.netloc + l
+                l = f"{parts.scheme}://{parts.netloc}{l}"
             if l.startswith("http"):
                 if self.same_schema and l.startswith(url.rstrip("/") + "/"):
                     out.add(l)
@@ -188,11 +190,9 @@ class HTTPFileSystem(AsyncFileSystem):
                 for u in out
             ]
         else:
-            return list(sorted(out))
-        return out
+            return sorted(out)
 
     async def _ls(self, url, detail=True, **kwargs):
-
         if self.use_listings_cache and url in self.dircache:
             out = self.dircache[url]
         else:
@@ -235,7 +235,7 @@ class HTTPFileSystem(AsyncFileSystem):
         kw.update(kwargs)
         logger.debug(rpath)
         session = await self.set_session()
-        async with session.get(self.encode_url(rpath), **self.kwargs) as r:
+        async with session.get(self.encode_url(rpath), **kw) as r:
             try:
                 size = int(r.headers["content-length"])
             except (ValueError, KeyError):
@@ -243,13 +243,20 @@ class HTTPFileSystem(AsyncFileSystem):
 
             callback.set_size(size)
             self._raise_not_found_for_status(r, rpath)
-            if not isfilelike(lpath):
-                lpath = open(lpath, "wb")
-            chunk = True
-            while chunk:
-                chunk = await r.content.read(chunk_size)
-                lpath.write(chunk)
-                callback.relative_update(len(chunk))
+            if isfilelike(lpath):
+                outfile = lpath
+            else:
+                outfile = open(lpath, "wb")
+
+            try:
+                chunk = True
+                while chunk:
+                    chunk = await r.content.read(chunk_size)
+                    outfile.write(chunk)
+                    callback.relative_update(len(chunk))
+            finally:
+                if not isfilelike(lpath):
+                    outfile.close()
 
     async def _put_file(
         self,
@@ -348,7 +355,7 @@ class HTTPFileSystem(AsyncFileSystem):
         if block_size and size:
             return HTTPFile(
                 self,
-                self.encode_url(path),
+                path,
                 session=session,
                 block_size=block_size,
                 mode=mode,
@@ -361,7 +368,7 @@ class HTTPFileSystem(AsyncFileSystem):
         else:
             return HTTPStreamFile(
                 self,
-                self.encode_url(path),
+                path,
                 mode=mode,
                 loop=self.loop,
                 session=session,
@@ -377,7 +384,7 @@ class HTTPFileSystem(AsyncFileSystem):
                 pass
         return AsyncStreamFile(
             self,
-            self.encode_url(path),
+            path,
             loop=self.loop,
             session=session,
             size=size,
@@ -422,7 +429,7 @@ class HTTPFileSystem(AsyncFileSystem):
 
         return {"name": url, "size": None, **info, "type": "file"}
 
-    async def _glob(self, path, **kwargs):
+    async def _glob(self, path, maxdepth=None, **kwargs):
         """
         Find files by glob-matching.
 
@@ -430,23 +437,21 @@ class HTTPFileSystem(AsyncFileSystem):
         but "?" is not considered as a character for globbing, because it is
         so common in URLs, often identifying the "query" part.
         """
+        if maxdepth is not None and maxdepth < 1:
+            raise ValueError("maxdepth must be at least 1")
         import re
 
         ends = path.endswith("/")
         path = self._strip_protocol(path)
-        indstar = path.find("*") if path.find("*") >= 0 else len(path)
-        indbrace = path.find("[") if path.find("[") >= 0 else len(path)
+        idx_star = path.find("*") if path.find("*") >= 0 else len(path)
+        idx_brace = path.find("[") if path.find("[") >= 0 else len(path)
 
-        ind = min(indstar, indbrace)
+        min_idx = min(idx_star, idx_brace)
 
         detail = kwargs.pop("detail", False)
 
         if not has_magic(path):
-            root = path
-            depth = 1
-            if ends:
-                path += "/*"
-            elif await self._exists(path):
+            if await self._exists(path):
                 if not detail:
                     return [path]
                 else:
@@ -456,13 +461,21 @@ class HTTPFileSystem(AsyncFileSystem):
                     return []  # glob of non-existent returns empty
                 else:
                     return {}
-        elif "/" in path[:ind]:
-            ind2 = path[:ind].rindex("/")
-            root = path[: ind2 + 1]
-            depth = None if "**" in path else path[ind2 + 1 :].count("/") + 1
+        elif "/" in path[:min_idx]:
+            min_idx = path[:min_idx].rindex("/")
+            root = path[: min_idx + 1]
+            depth = path[min_idx + 1 :].count("/") + 1
         else:
             root = ""
-            depth = None if "**" in path else path[ind + 1 :].count("/") + 1
+            depth = path[min_idx + 1 :].count("/") + 1
+
+        if "**" in path:
+            if maxdepth is not None:
+                idx_double_stars = path.find("**")
+                depth_double_stars = path[idx_double_stars:].count("/") + 1
+                depth = depth - depth_double_stars + maxdepth
+            else:
+                depth = None
 
         allpaths = await self._find(
             root, maxdepth=depth, withdirs=True, detail=True, **kwargs
@@ -489,14 +502,23 @@ class HTTPFileSystem(AsyncFileSystem):
             )
             + "$"
         )
-        pattern = re.sub("[*]{2}", "=PLACEHOLDER=", pattern)
+        pattern = re.sub("/[*]{2}", "=SLASH_DOUBLE_STARS=", pattern)
+        pattern = re.sub("[*]{2}/?", "=DOUBLE_STARS=", pattern)
         pattern = re.sub("[*]", "[^/]*", pattern)
-        pattern = re.compile(pattern.replace("=PLACEHOLDER=", ".*"))
+        pattern = re.sub("=SLASH_DOUBLE_STARS=", "(|/.*)", pattern)
+        pattern = re.sub("=DOUBLE_STARS=", ".*", pattern)
+        pattern = re.compile(pattern)
         out = {
             p: allpaths[p]
             for p in sorted(allpaths)
             if pattern.match(p.replace("//", "/").rstrip("/"))
         }
+
+        # Return directories only when the glob end by a slash
+        # This is needed for posix glob compliance
+        if ends:
+            out = {k: v for k, v in out.items() if v["type"] == "directory"}
+
         if detail:
             return out
         else:
@@ -597,7 +619,7 @@ class HTTPFile(AbstractBufferedFile):
         """
         logger.debug(f"Fetch all for {self}")
         if not isinstance(self.cache, AllBytes):
-            r = await self.session.get(self.url, **self.kwargs)
+            r = await self.session.get(self.fs.encode_url(self.url), **self.kwargs)
             async with r:
                 r.raise_for_status()
                 out = await r.read()
@@ -608,6 +630,20 @@ class HTTPFile(AbstractBufferedFile):
 
     _fetch_all = sync_wrapper(async_fetch_all)
 
+    def _parse_content_range(self, headers):
+        """Parse the Content-Range header"""
+        s = headers.get("Content-Range", "")
+        m = re.match(r"bytes (\d+-\d+|\*)/(\d+|\*)", s)
+        if not m:
+            return None, None, None
+
+        if m[1] == "*":
+            start = end = None
+        else:
+            start, end = [int(x) for x in m[1].split("-")]
+        total = None if m[2] == "*" else int(m[2])
+        return start, end, total
+
     async def async_fetch_range(self, start, end):
         """Download a block of data
 
@@ -619,20 +655,39 @@ class HTTPFile(AbstractBufferedFile):
         logger.debug(f"Fetch range for {self}: {start}-{end}")
         kwargs = self.kwargs.copy()
         headers = kwargs.pop("headers", {}).copy()
-        headers["Range"] = "bytes=%i-%i" % (start, end - 1)
-        logger.debug(str(self.url) + " : " + headers["Range"])
-        r = await self.session.get(self.url, headers=headers, **kwargs)
+        headers["Range"] = f"bytes={start}-{end - 1}"
+        logger.debug(f"{self.url} : {headers['Range']}")
+        r = await self.session.get(
+            self.fs.encode_url(self.url), headers=headers, **kwargs
+        )
         async with r:
             if r.status == 416:
                 # range request outside file
                 return b""
             r.raise_for_status()
-            if r.status == 206:
+
+            # If the server has handled the range request, it should reply
+            # with status 206 (partial content). But we'll guess that a suitable
+            # Content-Range header or a Content-Length no more than the
+            # requested range also mean we have got the desired range.
+            response_is_range = (
+                r.status == 206
+                or self._parse_content_range(r.headers)[0] == start
+                or int(r.headers.get("Content-Length", end + 1)) <= end - start
+            )
+
+            if response_is_range:
                 # partial content, as expected
                 out = await r.read()
-            elif int(r.headers.get("Content-Length", 0)) <= end - start:
-                out = await r.read()
+            elif start > 0:
+                raise ValueError(
+                    "The HTTP server doesn't appear to support range requests. "
+                    "Only reading this file from the beginning is supported. "
+                    "Open with block_size=0 for a streaming file interface."
+                )
             else:
+                # Response is not a range, but we want the start of the file,
+                # so we can read the required amount anyway.
                 cl = 0
                 out = []
                 while True:
@@ -690,13 +745,17 @@ class HTTPStreamFile(AbstractBufferedFile):
         super().__init__(fs=fs, path=url, mode=mode, cache_type="none", **kwargs)
 
         async def cor():
-            r = await self.session.get(url, **kwargs).__aenter__()
+            r = await self.session.get(self.fs.encode_url(url), **kwargs).__aenter__()
             self.fs._raise_not_found_for_status(r, url)
             return r
 
         self.r = sync(self.loop, cor)
 
-    def seek(self, *args, **kwargs):
+    def seek(self, loc, whence=0):
+        if loc == 0 and whence == 1:
+            return
+        if loc == self.loc and whence == 0:
+            return
         raise ValueError("Cannot seek streaming HTTP file")
 
     async def _read(self, num=-1):
@@ -733,7 +792,9 @@ class AsyncStreamFile(AbstractAsyncStreamedFile):
 
     async def read(self, num=-1):
         if self.r is None:
-            r = await self.session.get(self.url, **self.kwargs).__aenter__()
+            r = await self.session.get(
+                self.fs.encode_url(self.url), **self.kwargs
+            ).__aenter__()
             self.fs._raise_not_found_for_status(r, self.url)
             self.r = r
         out = await self.r.content.read(num)
@@ -751,7 +812,7 @@ async def get_range(session, url, start, end, file=None, **kwargs):
     # explicit get a range when we know it must be safe
     kwargs = kwargs.copy()
     headers = kwargs.pop("headers", {}).copy()
-    headers["Range"] = "bytes=%i-%i" % (start, end - 1)
+    headers["Range"] = f"bytes={start}-{end - 1}"
     r = await session.get(url, headers=headers, **kwargs)
     r.raise_for_status()
     async with r:
@@ -770,7 +831,7 @@ async def _file_info(url, session, size_policy="head", **kwargs):
     Default operation is to explicitly allow redirects and use encoding
     'identity' (no compression) to get the true size of the target.
     """
-    logger.debug("Retrieve file size for %s" % url)
+    logger.debug("Retrieve file size for %s", url)
     kwargs = kwargs.copy()
     ar = kwargs.pop("allow_redirects", True)
     head = kwargs.get("headers", {}).copy()
@@ -783,7 +844,7 @@ async def _file_info(url, session, size_policy="head", **kwargs):
     elif size_policy == "get":
         r = await session.get(url, allow_redirects=ar, **kwargs)
     else:
-        raise TypeError('size_policy must be "head" or "get", got %s' "" % size_policy)
+        raise TypeError(f'size_policy must be "head" or "get", got {size_policy}')
     async with r:
         r.raise_for_status()
 
@@ -792,7 +853,10 @@ async def _file_info(url, session, size_policy="head", **kwargs):
         #                 or 'Accept-Ranges': 'none' (not 'bytes')
         #  to mean streaming only, no random access => return None
         if "Content-Length" in r.headers:
-            info["size"] = int(r.headers["Content-Length"])
+            # Some servers may choose to ignore Accept-Encoding and return
+            # compressed content, in which case the returned size is unreliable.
+            if r.headers.get("Content-Encoding", "identity") == "identity":
+                info["size"] = int(r.headers["Content-Length"])
         elif "Content-Range" in r.headers:
             info["size"] = int(r.headers["Content-Range"].split("/")[1])
 
