@@ -3,15 +3,18 @@ import collections
 import io
 import itertools
 import logging
+import math
 import os
 from functools import lru_cache
+from typing import TYPE_CHECKING
 
 import fsspec.core
 
 try:
     import ujson as json
 except ImportError:
-    import json
+    if not TYPE_CHECKING:
+        import json
 
 from ..asyn import AsyncFileSystem
 from ..callbacks import _DEFAULT_CALLBACK
@@ -58,7 +61,7 @@ class RefsValuesView(collections.abc.ValuesView):
         yield from self._mapping._items.values()
         for field in self._mapping.listdir():
             chunk_sizes = self._mapping._get_chunk_sizes(field)
-            if chunk_sizes.size == 0:
+            if len(chunk_sizes) == 0:
                 yield self._mapping[field + "/0"]
                 continue
             yield from self._mapping._generate_all_records(field)
@@ -69,9 +72,22 @@ class RefsItemsView(collections.abc.ItemsView):
         return zip(self._mapping.keys(), self._mapping.values())
 
 
+def ravel_multi_index(idx, sizes):
+    val = 0
+    mult = 1
+    for i, s in zip(idx[::-1], sizes[::-1]):
+        val += i * mult
+        mult *= s
+    return val
+
+
 class LazyReferenceMapper(collections.abc.MutableMapping):
-    """Interface to read parquet store as if it were a standard kerchunk
-    references dict."""
+    """This interface can be used to read/write references from Parquet stores.
+    It is not intended for other types of references.
+    It can be used with Kerchunk's MultiZarrToZarr method to combine
+    references into a parquet store.
+    Examples of this use-case can be found here:
+    https://fsspec.github.io/kerchunk/advanced.html?highlight=parquet#parquet-storage"""
 
     # import is class level to prevent numpy dep requirement for fsspec
     @property
@@ -87,10 +103,7 @@ class LazyReferenceMapper(collections.abc.MutableMapping):
         return pd
 
     def __init__(
-        self,
-        root,
-        fs=None,
-        cache_size=128,
+        self, root, fs=None, out_root=None, cache_size=128, categorical_threshold=10
     ):
         """
         Parameters
@@ -99,62 +112,145 @@ class LazyReferenceMapper(collections.abc.MutableMapping):
             Root of parquet store
         fs : fsspec.AbstractFileSystem
             fsspec filesystem object, default is local filesystem.
-        cache_size : int
+        cache_size : int, default=128
             Maximum size of LRU cache, where cache_size*record_size denotes
             the total number of references that can be loaded in memory at once.
+        categorical_threshold : int
+            Encode urls as pandas.Categorical to reduce memory footprint if the ratio
+            of the number of unique urls to total number of refs for each variable
+            is greater than or equal to this number. (default 10)
+
+
         """
         self.root = root
         self.chunk_sizes = {}
         self._items = {}
         self.dirs = None
         self.fs = fsspec.filesystem("file") if fs is None else fs
-        with self.fs.open("/".join([self.root, ".zmetadata"]), "rb") as f:
-            self._items[".zmetadata"] = f.read()
+        self._items[".zmetadata"] = self.fs.cat_file(
+            "/".join([self.root, ".zmetadata"])
+        )
         met = json.loads(self._items[".zmetadata"])
         self.record_size = met["record_size"]
         self.zmetadata = met["metadata"]
-        if self.root:
-            self.url = self.root + "/{field}/refs.{record}.parq"
-        else:
-            self.url = "{field}/refs.{record}.parq"
+        self.url = self.root + "/{field}/refs.{record}.parq"
+        self.out_root = out_root or self.root
+        self.cat_thresh = categorical_threshold
 
         # Define function to open and decompress refs
         @lru_cache(maxsize=cache_size)
         def open_refs(field, record):
+            """cached parquet file loader"""
             path = self.url.format(field=field, record=record)
-            with self.fs.open(path) as f:
-                df = self.pd.read_parquet(f, engine="fastparquet")
+            data = io.BytesIO(self.fs.cat_file(path))
+            df = self.pd.read_parquet(data, engine="fastparquet")
             refs = {c: df[c].values for c in df.columns}
-            # Return both df and dict of views because the former is
-            # more convenient for iterating sequentially while the latter
-            # is faster for random access.
-            return df, refs
+            return refs
 
         self.open_refs = open_refs
 
+    @staticmethod
+    def create(record_size, root, fs, **kwargs):
+        met = {"metadata": {}, "record_size": record_size}
+        fs.pipe("/".join([root, ".zmetadata"]), json.dumps(met).encode())
+        return LazyReferenceMapper(root, fs, **kwargs)
+
     def listdir(self, basename=True):
+        """List top-level directories"""
         if self.dirs is None:
             dirs = [p.split("/", 1)[0] for p in self.zmetadata]
-            self.dirs = set(sorted(p for p in dirs if p and not p.startswith(".")))
+            self.dirs = {p for p in dirs if p and not p.startswith(".")}
         listing = self.dirs
         if basename:
             listing = [os.path.basename(path) for path in listing]
         return listing
 
+    def ls(self, path="", detail=True):
+        """Shortcut file listings"""
+        if not path:
+            dirnames = self.listdir()
+            others = set(
+                [".zmetadata"]
+                + [name for name in self.zmetadata if "/" not in name]
+                + [name for name in self._items if "/" not in name]
+            )
+            if detail is False:
+                others.update(dirnames)
+                return sorted(others)
+            dirinfo = [
+                {"name": name, "type": "directory", "size": 0} for name in dirnames
+            ]
+            fileinfo = [
+                {
+                    "name": name,
+                    "type": "file",
+                    "size": len(
+                        json.dumps(self.zmetadata[name])
+                        if name in self.zmetadata
+                        else self._items[name]
+                    ),
+                }
+                for name in others
+            ]
+            return sorted(dirinfo + fileinfo, key=lambda s: s["name"])
+        parts = path.split("/", 1)
+        if len(parts) > 1:
+            raise FileNotFoundError("Cannot list within directories right now")
+        field = parts[0]
+        others = set(
+            [name for name in self.zmetadata if name.startswith(f"{path}/")]
+            + [name for name in self._items if name.startswith(f"{path}/")]
+        )
+        fileinfo = [
+            {
+                "name": name,
+                "type": "file",
+                "size": len(
+                    json.dumps(self.zmetadata[name])
+                    if name in self.zmetadata
+                    else self._items[name]
+                ),
+            }
+            for name in others
+        ]
+        keys = self._keys_in_field(field)
+
+        if detail is False:
+            return list(others) + list(keys)
+        recs = self._generate_all_records(field)
+        recinfo = [
+            {"name": name, "type": "file", "size": rec[-1]}
+            for name, rec in zip(keys, recs)
+            if rec[0]  # filters out path==None, deleted/missing
+        ]
+        return fileinfo + recinfo
+
     def _load_one_key(self, key):
+        """Get the reference for one key
+
+        Returns bytes, one-element list or three-element list.
+        """
         if key in self._items:
             return self._items[key]
         elif key in self.zmetadata:
             return json.dumps(self.zmetadata[key]).encode()
-        elif "/" not in key:
+        elif "/" not in key or self._is_meta(key):
             raise KeyError(key)
         field, sub_key = key.split("/")
+        record, _, _ = self._key_to_record(key)
+        maybe = self._items.get((field, key), {}).get(sub_key, False)
+        if maybe is None:
+            # explicitly deleted
+            raise KeyError
+        elif maybe:
+            return maybe
+
         # Chunk keys can be loaded from row group and cached in LRU cache
         try:
             record, ri, chunk_size = self._key_to_record(key)
             if chunk_size == 0:
                 return b""
-            _, refs = self.open_refs(field, record)
+            refs = self.open_refs(field, record)
         except (ValueError, TypeError, FileNotFoundError):
             raise KeyError(key)
         columns = ["path", "offset", "size", "raw"]
@@ -162,47 +258,57 @@ class LazyReferenceMapper(collections.abc.MutableMapping):
         raw = selection[-1]
         if raw is not None:
             return raw
-        data = selection[:-1]
-        if data[1:] == [0, 0]:
-            data = data[:1]
-        return data
+        if selection[0] is None:
+            raise KeyError("This reference has been deleted")
+        if selection[1:3] == [0, 0]:
+            # URL only
+            return selection[:1]
+        # URL, offset, size
+        return selection[:3]
 
+    @lru_cache(4096)
     def _key_to_record(self, key):
+        """Details needed to construct a reference for one key"""
         field, chunk = key.split("/")
         chunk_sizes = self._get_chunk_sizes(field)
-        if chunk_sizes.size == 0:
+        if len(chunk_sizes) == 0:
             return 0, 0, 0
-        chunk_idx = self.np.array([int(c) for c in chunk.split(".")])
-        chunk_number = self.np.ravel_multi_index(chunk_idx, chunk_sizes)
+        chunk_idx = [int(c) for c in chunk.split(".")]
+        chunk_number = ravel_multi_index(chunk_idx, chunk_sizes)
         record = chunk_number // self.record_size
         ri = chunk_number % self.record_size
-        return record, ri, chunk_sizes.size
+        return record, ri, len(chunk_sizes)
 
     def _get_chunk_sizes(self, field):
+        """The number of chunks along each axis for a given field"""
         if field not in self.chunk_sizes:
             zarray = self.zmetadata[f"{field}/.zarray"]
-            size_ratio = self.np.array(zarray["shape"]) / self.np.array(
-                zarray["chunks"]
-            )
-            self.chunk_sizes[field] = self.np.ceil(size_ratio).astype(int)
+            size_ratio = [
+                math.ceil(s / c) for s, c in zip(zarray["shape"], zarray["chunks"])
+            ]
+            self.chunk_sizes[field] = size_ratio
         return self.chunk_sizes[field]
 
     def _generate_record(self, field, record):
-        df, _ = self.open_refs(field, record)
-        it = df.itertuples(name=None, index=False)
-        if df.columns.size == 3:
+        """The references for a given parquet file of a given field"""
+        refs = self.open_refs(field, record)
+        it = iter(zip(refs.values()))
+        if len(refs) == 3:
             # All urls
             return (list(t) for t in it)
-        elif df.columns.size == 1:
+        elif len(refs) == 1:
             # All raws
-            return (t[0] for t in it)
+            return refs["raw"]
         else:
             # Mix of urls and raws
             return (list(t[:3]) if not t[3] else t[3] for t in it)
 
     def _generate_all_records(self, field):
-        chunk_size = self._get_chunk_sizes(field)
-        nrec = int(self.np.ceil(self.np.product(chunk_size) / self.record_size))
+        """Load all the references within a field by iterating over the parquet files"""
+        nrec = 1
+        for ch in self._get_chunk_sizes(field):
+            nrec *= ch
+        nrec = math.ceil(nrec / self.record_size)
         for record in range(nrec):
             yield from self._generate_record(field, record)
 
@@ -212,18 +318,31 @@ class LazyReferenceMapper(collections.abc.MutableMapping):
     def items(self):
         return RefsItemsView(self)
 
+    def __hash__(self):
+        return id(self)
+
+    @lru_cache(20)
     def __getitem__(self, key):
-        if key in self._items:
-            val = self._items[key]
-            if val is None:
-                raise KeyError
-        if key in self.zmetadata:
-            # spec requires bytes even if we already decoded the metadata
-            return json.dumps(self.zmetadata[key]).encode()
         return self._load_one_key(key)
 
     def __setitem__(self, key, value):
-        self._items[key] = value
+        if "/" in key and not self._is_meta(key):
+            field, chunk = key.split("/")
+            record, i, _ = self._key_to_record(key)
+            subdict = self._items.setdefault((field, record), {})
+            subdict[i] = value
+            if len(subdict) == self.record_size:
+                self.write(field, record)
+        else:
+            # metadata or top-level
+            self._items[key] = value
+            self.zmetadata[key] = json.loads(
+                value.decode() if isinstance(value, bytes) else value
+            )
+
+    @staticmethod
+    def _is_meta(key):
+        return key.startswith(".z") or "/.z" in key
 
     def __delitem__(self, key):
         if key in self._items:
@@ -231,9 +350,107 @@ class LazyReferenceMapper(collections.abc.MutableMapping):
         elif key in self.zmetadata:
             del self.zmetadata[key]
         else:
-            # TODO: Add method to RefFs that tracks changes in _items
-            # and updates / copies appropriate parquet ref files.
-            self._items[key] = None
+            if "/" in key and not self._is_meta(key):
+                field, chunk = key.split("/")
+                record, _, _ = self._key_to_record(key)
+                subdict = self._items.setdefault((field, record), {})
+                subdict[chunk] = None
+                if len(subdict) == self.record_size:
+                    self.write(field, record)
+            else:
+                # metadata or top-level
+                self._items[key] = None
+
+    def write(self, field, record, base_url=None, storage_options=None):
+        # extra requirements if writing
+        import kerchunk.df
+        import numpy as np
+        import pandas as pd
+
+        # TODO: if the dict is incomplete, also load records and merge in
+        partition = self._items[(field, record)]
+        fn = f"{base_url or self.out_root}/{field}/refs.{record}.parq"
+
+        ####
+        paths = np.full(self.record_size, np.nan, dtype="O")
+        offsets = np.zeros(self.record_size, dtype="int64")
+        sizes = np.zeros(self.record_size, dtype="int64")
+        raws = np.full(self.record_size, np.nan, dtype="O")
+        nraw = 0
+        npath = 0
+        for j, data in partition.items():
+            if isinstance(data, list):
+                npath += 1
+                paths[j] = data[0]
+                if len(data) > 1:
+                    offsets[j] = data[1]
+                    sizes[j] = data[2]
+            else:
+                nraw += 1
+                raws[j] = kerchunk.df._proc_raw(data)
+        # TODO: only save needed columns
+        df = pd.DataFrame(
+            {
+                "path": paths,
+                "offset": offsets,
+                "size": sizes,
+                "raw": raws,
+            },
+            copy=False,
+        )
+        if df.path.count() / (df.path.nunique() or 1) > self.cat_thresh:
+            df["path"] = df["path"].astype("category")
+        object_encoding = {"raw": "bytes", "path": "utf8"}
+        has_nulls = ["path", "raw"]
+
+        self.fs.mkdirs(f"{base_url or self.out_root}/{field}", exist_ok=True)
+        df.to_parquet(
+            fn,
+            engine="fastparquet",
+            storage_options=storage_options
+            or getattr(self.fs, "storage_options", None),
+            compression="zstd",
+            index=False,
+            stats=False,
+            object_encoding=object_encoding,
+            has_nulls=has_nulls,
+            # **kwargs,
+        )
+        partition.clear()
+        self._items.pop((field, record))
+
+    def flush(self, base_url=None, storage_options=None):
+        """Output any modified or deleted keys
+
+        Parameters
+        ----------
+        base_url: str
+            Location of the output
+        """
+        # write what we have so far and clear sub chunks
+        for thing in list(self._items):
+            if isinstance(thing, tuple):
+                field, record = thing
+                self.write(
+                    field,
+                    record,
+                    base_url=base_url,
+                    storage_options=storage_options,
+                )
+
+        # gather .zmetadata from self._items and write that too
+        for k in list(self._items):
+            if k != ".zmetadata" and ".z" in k:
+                self.zmetadata[k] = json.loads(self._items.pop(k))
+        met = {"metadata": self.zmetadata, "record_size": self.record_size}
+        self._items[".zmetadata"] = json.dumps(met).encode()
+        self.fs.pipe(
+            "/".join([base_url or self.out_root, ".zmetadata"]),
+            self._items[".zmetadata"],
+        )
+
+        # TODO: only clear those that we wrote to?
+        self.open_refs.cache_clear()
 
     def __len__(self):
         # Caveat: This counts expected references, not actual
@@ -252,16 +469,33 @@ class LazyReferenceMapper(collections.abc.MutableMapping):
     def __iter__(self):
         # Caveat: Note that this generates all expected keys, but does not
         # account for reference keys that are missing.
-        yield from self.zmetadata
-        yield from self._items
+        metas = set(self.zmetadata)
+        metas.update(self._items)
+        for bit in metas:
+            if isinstance(bit, str):
+                yield bit
         for field in self.listdir():
-            chunk_sizes = self._get_chunk_sizes(field)
-            if chunk_sizes.size == 0:
-                yield field + "/0"
-                continue
-            inds = self.np.ndindex(*chunk_sizes)
-            for ind in inds:
-                yield field + "/" + ".".join([str(c) for c in ind])
+            yield from self._keys_in_field(field)
+
+    def __contains__(self, item):
+        try:
+            self._load_one_key(item)
+            return True
+        except KeyError:
+            return False
+
+    def _keys_in_field(self, field):
+        """List key names in given field
+
+        Produces strings like "field/x.y" appropriate from the chunking of the array
+        """
+        chunk_sizes = self._get_chunk_sizes(field)
+        if len(chunk_sizes) == 0:
+            yield field + "/0"
+            return
+        inds = itertools.product(*(range(i) for i in chunk_sizes))
+        for ind in inds:
+            yield field + "/" + ".".join([str(c) for c in ind])
 
 
 class ReferenceFileSystem(AsyncFileSystem):
@@ -368,17 +602,18 @@ class ReferenceFileSystem(AsyncFileSystem):
             dic = dict(
                 **(ref_storage_args or target_options or {}), protocol=target_protocol
             )
-            ref_fs, fo = fsspec.core.url_to_fs(fo, **dic)
+            ref_fs, fo2 = fsspec.core.url_to_fs(fo, **dic)
             if ref_fs.isfile(fo):
                 # text JSON
-                with ref_fs.open(fo, "rb") as f:
+                with fsspec.open(fo, "rb", **dic) as f:
                     logger.info("Read reference from URL %s", fo)
                     text = json.load(f)
                 self._process_references(text, template_overrides)
             else:
                 # Lazy parquet refs
+                logger.info("Open lazy reference dict from URL %s", fo)
                 self.references = LazyReferenceMapper(
-                    fo,
+                    fo2,
                     fs=ref_fs,
                     cache_size=cache_size,
                 )
@@ -420,9 +655,11 @@ class ReferenceFileSystem(AsyncFileSystem):
                     ref = ref()
                 if isinstance(ref, list) and ref[0]:
                     protocol, _ = fsspec.core.split_protocol(ref[0])
-                    if protocol and protocol not in self.fss:
+                    if protocol not in self.fss:
                         fs = filesystem(protocol, **(remote_options or {}))
                         self.fss[protocol] = fs
+                        # only use first remote URL
+                        break
 
         if remote_protocol and remote_protocol not in self.fss:
             fs = filesystem(remote_protocol, **(remote_options or {}))
@@ -548,11 +785,11 @@ class ReferenceFileSystem(AsyncFileSystem):
                     urls.append(u)
                     starts.append(s)
                     ends.append(e)
-                except FileNotFoundError as e:
+                except FileNotFoundError as err:
                     if on_error == "raise":
                         raise
                     if on_error != "omit":
-                        out[p] = e
+                        out[p] = err
 
             # process references into form for merging
             urls2 = []
@@ -669,7 +906,6 @@ class ReferenceFileSystem(AsyncFileSystem):
         self.references.update(self._process_gen(references.get("gen", [])))
 
     def _process_templates(self, tmp):
-
         self.templates = {}
         if self.template_overrides is not None:
             tmp.update(self.template_overrides)
@@ -684,7 +920,6 @@ class ReferenceFileSystem(AsyncFileSystem):
                 self.templates[k] = v
 
     def _process_gen(self, gens):
-
         out = {}
         for gen in gens:
             dimension = {
@@ -747,6 +982,12 @@ class ReferenceFileSystem(AsyncFileSystem):
 
     def ls(self, path, detail=True, **kwargs):
         path = self._strip_protocol(path)
+        if isinstance(self.references, LazyReferenceMapper):
+            try:
+                return self.references.ls(path, detail)
+            except KeyError:
+                pass
+            raise FileNotFoundError(f"'{path}' is not a known key")
         if not self.dircache:
             self._dircache_from_items()
         out = self._ls_from_cache(path)
@@ -762,6 +1003,8 @@ class ReferenceFileSystem(AsyncFileSystem):
     def isdir(self, path):  # overwrite auto-sync version
         if self.dircache:
             return path in self.dircache
+        elif isinstance(self.references, LazyReferenceMapper):
+            return path in self.references.listdir("")
         else:
             # this may be faster than building dircache for single calls, but
             # by looping will be slow for many calls; could cache it?
@@ -774,7 +1017,6 @@ class ReferenceFileSystem(AsyncFileSystem):
         return self.ls(path, detail, **kwargs)
 
     def find(self, path, maxdepth=None, withdirs=False, detail=False, **kwargs):
-        # TODO: details
         if withdirs:
             return super().find(
                 path, maxdepth=maxdepth, withdirs=withdirs, detail=detail, **kwargs
@@ -792,8 +1034,8 @@ class ReferenceFileSystem(AsyncFileSystem):
             return r
 
     def info(self, path, **kwargs):
-        if path in self.references:
-            out = self.references[path]
+        out = self.references.get(path)
+        if out is not None:
             if isinstance(out, (str, bytes)):
                 # decode base64 here
                 return {"name": path, "type": "file", "size": len(out)}
