@@ -1,10 +1,17 @@
-import paramiko
-from stat import S_ISDIR, S_ISLNK
+import datetime
+import logging
+import os
 import types
 import uuid
+from stat import S_ISDIR, S_ISLNK
+
+import paramiko
+
 from .. import AbstractFileSystem
 from ..utils import infer_storage_options
 
+logger = logging.getLogger("fsspec.sftp")
+
 
 class SFTPFileSystem(AbstractFileSystem):
     """Files over SFTP/SSH
@@ -29,18 +36,19 @@ class SFTPFileSystem(AbstractFileSystem):
             Location on the server to put files, when within a transaction
         ssh_kwargs: dict
             Parameters passed on to connection. See details in
-            http://docs.paramiko.org/en/2.4/api/client.html#paramiko.client.SSHClient.connect
+            https://docs.paramiko.org/en/3.3/api/client.html#paramiko.client.SSHClient.connect
             May include port, username, password...
         """
         if self._cached:
             return
-        super(SFTPFileSystem, self).__init__(**ssh_kwargs)
-        self.temppath = ssh_kwargs.pop("temppath", "/tmp")
+        super().__init__(**ssh_kwargs)
+        self.temppath = ssh_kwargs.pop("temppath", "/tmp")  # remote temp directory
         self.host = host
         self.ssh_kwargs = ssh_kwargs
         self._connect()
 
     def _connect(self):
+        logger.debug("Connecting to SFTP server %s", self.host)
         self.client = paramiko.SSHClient()
         self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
         self.client.connect(self.host, **self.ssh_kwargs)
@@ -57,54 +65,80 @@ class SFTPFileSystem(AbstractFileSystem):
         out.pop("protocol", None)
         return out
 
-    def mkdir(self, path, mode=511):
-        self.ftp.mkdir(path, mode)
+    def mkdir(self, path, create_parents=False, mode=511):
+        logger.debug("Creating folder %s", path)
+        if self.exists(path):
+            raise FileExistsError(f"File exists: {path}")
+
+        if create_parents:
+            self.makedirs(path)
+        else:
+            self.ftp.mkdir(path, mode)
 
     def makedirs(self, path, exist_ok=False, mode=511):
         if self.exists(path) and not exist_ok:
-            raise FileExistsError("File exists: {}".format(path))
+            raise FileExistsError(f"File exists: {path}")
 
         parts = path.split("/")
         path = ""
 
         for part in parts:
-            path += "/" + part
+            path += f"/{part}"
             if not self.exists(path):
-                self.mkdir(path, mode)
+                self.ftp.mkdir(path, mode)
 
     def rmdir(self, path):
+        logger.debug("Removing folder %s", path)
         self.ftp.rmdir(path)
 
     def info(self, path):
-        s = self.ftp.stat(path)
-        if S_ISDIR(s.st_mode):
+        stat = self._decode_stat(self.ftp.stat(path))
+        stat["name"] = path
+        return stat
+
+    @staticmethod
+    def _decode_stat(stat, parent_path=None):
+        if S_ISDIR(stat.st_mode):
             t = "directory"
-        elif S_ISLNK(s.st_mode):
+        elif S_ISLNK(stat.st_mode):
             t = "link"
         else:
             t = "file"
-        return {
-            "name": path + "/" if t == "directory" else path,
-            "size": s.st_size,
+        out = {
+            "name": "",
+            "size": stat.st_size,
             "type": t,
-            "uid": s.st_uid,
-            "gui": s.st_gid,
-            "time": s.st_atime,
-            "mtime": s.st_mtime,
+            "uid": stat.st_uid,
+            "gid": stat.st_gid,
+            "time": datetime.datetime.fromtimestamp(
+                stat.st_atime, tz=datetime.timezone.utc
+            ),
+            "mtime": datetime.datetime.fromtimestamp(
+                stat.st_mtime, tz=datetime.timezone.utc
+            ),
         }
+        if parent_path:
+            out["name"] = "/".join([parent_path.rstrip("/"), stat.filename])
+        return out
 
     def ls(self, path, detail=False):
-        out = ["/".join([path.rstrip("/"), p]) for p in self.ftp.listdir(path)]
-        out = [self.info(o) for o in out]
+        logger.debug("Listing folder %s", path)
+        stats = [self._decode_stat(stat, path) for stat in self.ftp.listdir_iter(path)]
         if detail:
-            return out
-        return sorted([p["name"] for p in out])
+            return stats
+        else:
+            paths = [stat["name"] for stat in stats]
+            return sorted(paths)
 
-    def put(self, lpath, rpath):
+    def put(self, lpath, rpath, callback=None, **kwargs):
+        logger.debug("Put file %s into %s", lpath, rpath)
         self.ftp.put(lpath, rpath)
 
-    def get(self, rpath, lpath):
-        self.ftp.get(rpath, lpath)
+    def get_file(self, rpath, lpath, **kwargs):
+        if self.isdir(rpath):
+            os.makedirs(lpath, exist_ok=True)
+        else:
+            self.ftp.get(self._strip_protocol(rpath), lpath)
 
     def _open(self, path, mode="rb", block_size=None, **kwargs):
         """
@@ -112,9 +146,10 @@ class SFTPFileSystem(AbstractFileSystem):
             If 0, no buffering, if 1, line buffering, if >1, buffer that many
             bytes, if None use default from paramiko.
         """
+        logger.debug("Opening file %s", path)
         if kwargs.get("autocommit", True) is False:
             # writes to temporary file, move on commit
-            path2 = "{}/{}".format(self.temppath, uuid.uuid4())
+            path2 = "/".join([self.temppath, str(uuid.uuid4())])
             f = self.ftp.open(path2, mode, bufsize=block_size if block_size else -1)
             f.temppath = path2
             f.targetpath = path
@@ -132,6 +167,7 @@ class SFTPFileSystem(AbstractFileSystem):
             self.ftp.remove(path)
 
     def mv(self, old, new):
+        logger.debug("Renaming %s into %s", old, new)
         self.ftp.posix_rename(old, new)
 
 
