@@ -1,13 +1,16 @@
 import asyncio
 import io
+import json
 import os
 import sys
 import time
 
+import aiohttp
 import pytest
 
 import fsspec.asyn
 import fsspec.utils
+from fsspec.implementations.http import HTTPStreamFile
 from fsspec.tests.conftest import data, reset_files, server, win  # noqa: F401
 
 
@@ -200,6 +203,9 @@ def test_methods(server):
         {"give_range": "true"},
         {"give_length": "true", "head_not_auth": "true"},
         {"give_range": "true", "head_not_auth": "true"},
+        {"use_206": "true", "head_ok": "true", "head_give_length": "true"},
+        {"use_206": "true", "give_length": "true"},
+        {"use_206": "true", "give_range": "true"},
     ],
 )
 def test_random_access(server, headers):
@@ -219,6 +225,41 @@ def test_random_access(server, headers):
     assert f.closed
 
 
+@pytest.mark.parametrize(
+    "headers",
+    [
+        {"ignore_range": "true", "head_ok": "true", "head_give_length": "true"},
+        {"ignore_range": "true", "give_length": "true"},
+        {"ignore_range": "true", "give_range": "true"},
+    ],
+)
+def test_no_range_support(server, headers):
+    h = fsspec.filesystem("http", headers=headers)
+    url = server + "/index/realfile"
+    with h.open(url, "rb") as f:
+        # Random access is not possible if the server doesn't respect Range
+        f.seek(5)
+        with pytest.raises(ValueError):
+            f.read(10)
+
+        # Reading from the beginning should still work
+        f.seek(0)
+        assert f.read(10) == data[:10]
+
+
+def test_stream_seek(server):
+    h = fsspec.filesystem("http")
+    url = server + "/index/realfile"
+    with h.open(url, "rb") as f:
+        f.seek(0)  # is OK
+        data1 = f.read(5)
+        assert len(data1) == 5
+        f.seek(5)
+        f.seek(0, 1)
+        data2 = f.read()
+        assert data1 + data2 == data
+
+
 def test_mapper_url(server):
     h = fsspec.filesystem("http")
     mapper = h.get_mapper(server + "/index/")
@@ -240,6 +281,18 @@ def test_content_length_zero(server):
         assert f.read() == data
 
 
+def test_content_encoding_gzip(server):
+    h = fsspec.filesystem(
+        "http", headers={"give_length": "true", "gzip_encoding": "true"}
+    )
+    url = server + "/index/realfile"
+
+    with h.open(url, "rb") as f:
+        assert isinstance(f, HTTPStreamFile)
+        assert f.size is None
+        assert f.read() == data
+
+
 def test_download(server, tmpdir):
     h = fsspec.filesystem("http", headers={"give_length": "true", "head_ok": "true "})
     url = server + "/index/realfile"
@@ -289,6 +342,15 @@ def test_cat_file_range(server):
     assert h.cat(urla, end=-10) == data[:-10]
 
 
+def test_cat_file_range_numpy(server):
+    np = pytest.importorskip("numpy")
+    h = fsspec.filesystem("http", headers={"give_length": "true", "head_ok": "true "})
+    urla = server + "/index/realfile"
+    assert h.cat(urla, start=np.int8(1), end=np.int8(10)) == data[1:10]
+    out = h.cat_ranges([urla, urla], starts=np.array([1, 5]), ends=np.array([10, 15]))
+    assert out == [data[1:10], data[5:15]]
+
+
 def test_mcat_cache(server):
     urla = server + "/index/realfile"
     urlb = server + "/index/otherfile"
@@ -438,3 +500,61 @@ async def test_async_file(server):
         out2 = await f.read()
         assert data == out1 + out2
     await fs._session.close()
+
+
+def test_encoded(server):
+    fs = fsspec.filesystem("http", encoded=True)
+    out = fs.cat(server + "/Hello%3A%20G%C3%BCnter", headers={"give_path": "true"})
+    assert json.loads(out)["path"] == "/Hello%3A%20G%C3%BCnter"
+    with pytest.raises(aiohttp.client_exceptions.ClientError):
+        fs.cat(server + "/Hello: Günter", headers={"give_path": "true"})
+
+    fs = fsspec.filesystem("http", encoded=False)
+    out = fs.cat(server + "/Hello: Günter", headers={"give_path": "true"})
+    assert json.loads(out)["path"] == "/Hello:%20G%C3%BCnter"
+
+
+def test_with_cache(server):
+    fs = fsspec.filesystem("http", headers={"head_ok": "true", "give_length": "true"})
+    fn = server + "/index/realfile"
+    fs1 = fsspec.filesystem("blockcache", fs=fs)
+    with fs1.open(fn, "rb") as f:
+        out = f.read()
+    assert out == fs1.cat(fn)
+
+
+@pytest.mark.asyncio
+async def test_async_expand_path(server):
+    fs = fsspec.filesystem("http", asynchronous=True, skip_instance_cache=True)
+
+    # maxdepth=1
+    assert await fs._expand_path(server + "/index", recursive=True, maxdepth=1) == [
+        server + "/index",
+        server + "/index/realfile",
+    ]
+
+    # maxdepth=0
+    with pytest.raises(ValueError):
+        await fs._expand_path(server + "/index", maxdepth=0)
+    with pytest.raises(ValueError):
+        await fs._expand_path(server + "/index", recursive=True, maxdepth=0)
+
+    await fs._session.close()
+
+
+@pytest.mark.asyncio
+async def test_async_walk(server):
+    fs = fsspec.filesystem("http", asynchronous=True, skip_instance_cache=True)
+
+    # No maxdepth
+    res = []
+    async for a in fs._walk(server + "/index"):
+        res.append(a)
+    assert res == [(server + "/index", [], ["realfile"])]
+
+    # maxdepth=0
+    with pytest.raises(ValueError):
+        async for a in fs._walk(server + "/index", maxdepth=0):
+            pass
+
+    await fs._session.close()
