@@ -1,5 +1,3 @@
-from __future__ import absolute_import, division, print_function
-
 import bz2
 import gzip
 import os
@@ -58,7 +56,7 @@ def filetexts(d, open=open, mode="t"):
     try:
         os.chdir(dirname)
         for filename, text in d.items():
-            f = open(filename, "w" + mode)
+            f = open(filename, f"w{mode}")
             try:
                 f.write(text)
             finally:
@@ -73,7 +71,7 @@ def filetexts(d, open=open, mode="t"):
             if os.path.exists(filename):
                 try:
                     os.remove(filename)
-                except (IOError, OSError):
+                except OSError:
                     pass
     finally:
         os.chdir(odir)
@@ -81,7 +79,7 @@ def filetexts(d, open=open, mode="t"):
 
 def test_urlpath_inference_strips_protocol(tmpdir):
     tmpdir = make_path_posix(str(tmpdir))
-    paths = ["/".join([tmpdir, "test.%02d.csv" % i]) for i in range(20)]
+    paths = ["/".join([tmpdir, f"test.{i:02d}.csv"]) for i in range(20)]
 
     for path in paths:
         with open(path, "wb") as f:
@@ -104,10 +102,11 @@ def test_urlpath_inference_errors():
         get_fs_token_paths([])
     assert "empty" in str(err.value)
 
+    pytest.importorskip("s3fs")
     # Protocols differ
     with pytest.raises(ValueError) as err:
         get_fs_token_paths(["s3://test/path.csv", "/other/path.csv"])
-    assert "protocol" in str(err.value)
+    assert "Protocol" in str(err.value)
 
 
 def test_urlpath_expand_read():
@@ -144,11 +143,11 @@ def test_urlpath_expand_write():
     """Make sure * is expanded in file paths when writing."""
     _, _, paths = get_fs_token_paths("prefix-*.csv", mode="wb", num=2)
     assert all(
-        [p.endswith(pa) for p, pa in zip(paths, ["/prefix-0.csv", "/prefix-1.csv"])]
+        p.endswith(pa) for p, pa in zip(paths, ["/prefix-0.csv", "/prefix-1.csv"])
     )
     _, _, paths = get_fs_token_paths(["prefix-*.csv"], mode="wb", num=2)
     assert all(
-        [p.endswith(pa) for p, pa in zip(paths, ["/prefix-0.csv", "/prefix-1.csv"])]
+        p.endswith(pa) for p, pa in zip(paths, ["/prefix-0.csv", "/prefix-1.csv"])
     )
     # we can read with multiple masks, but not write
     with pytest.raises(ValueError):
@@ -222,7 +221,7 @@ def test_isfile():
     with filetexts(files, mode="b"):
         for f in files.keys():
             assert fs.isfile(f)
-            assert fs.isfile("file://" + f)
+            assert fs.isfile(f"file://{f}")
         assert not fs.isfile("not-a-file")
         assert not fs.isfile("file://not-a-file")
 
@@ -294,7 +293,7 @@ def test_abs_paths(tmpdir):
 def test_glob_weird_characters(tmpdir, sep, chars):
     tmpdir = make_path_posix(str(tmpdir))
 
-    subdir = tmpdir + sep + "test" + chars + "x"
+    subdir = f"{tmpdir}{sep}test{chars}x"
     try:
         os.makedirs(subdir, exist_ok=True)
     except OSError as e:
@@ -322,7 +321,9 @@ def test_globfind_dirs(tmpdir):
         fs.glob(tmpdir + "/dir/*", detail=True)[tmpdir + "/dir/afile"]["type"] == "file"
     )
     assert [tmpdir + "/dir/afile"] == fs.find(tmpdir)
-    assert [tmpdir + "/dir", tmpdir + "/dir/afile"] == fs.find(tmpdir, withdirs=True)
+    assert [tmpdir, tmpdir + "/dir", tmpdir + "/dir/afile"] == fs.find(
+        tmpdir, withdirs=True
+    )
 
 
 def test_touch(tmpdir):
@@ -365,51 +366,55 @@ def test_directories(tmpdir):
     assert fs.ls(fs.root_marker)
 
 
-def test_file_ops(tmpdir):
+@pytest.mark.parametrize("file_protocol", ["", "file://"])
+def test_file_ops(tmpdir, file_protocol):
     tmpdir = make_path_posix(str(tmpdir))
+    tmpdir_with_protocol = file_protocol + tmpdir
     fs = LocalFileSystem(auto_mkdir=True)
     with pytest.raises(FileNotFoundError):
-        fs.info(tmpdir + "/nofile")
-    fs.touch(tmpdir + "/afile")
-    i1 = fs.ukey(tmpdir + "/afile")
+        fs.info(tmpdir_with_protocol + "/nofile")
+    fs.touch(tmpdir_with_protocol + "/afile")
+    i1 = fs.ukey(tmpdir_with_protocol + "/afile")
 
-    assert tmpdir + "/afile" in fs.ls(tmpdir)
+    assert tmpdir + "/afile" in fs.ls(tmpdir_with_protocol)
 
-    with fs.open(tmpdir + "/afile", "wb") as f:
+    with fs.open(tmpdir_with_protocol + "/afile", "wb") as f:
         f.write(b"data")
-    i2 = fs.ukey(tmpdir + "/afile")
+    i2 = fs.ukey(tmpdir_with_protocol + "/afile")
     assert i1 != i2  # because file changed
 
-    fs.copy(tmpdir + "/afile", tmpdir + "/afile2")
-    assert tmpdir + "/afile2" in fs.ls(tmpdir)
+    fs.copy(tmpdir_with_protocol + "/afile", tmpdir_with_protocol + "/afile2")
+    assert tmpdir + "/afile2" in fs.ls(tmpdir_with_protocol)
 
-    fs.move(tmpdir + "/afile", tmpdir + "/afile3")
-    assert not fs.exists(tmpdir + "/afile")
+    fs.move(tmpdir_with_protocol + "/afile", tmpdir_with_protocol + "/afile3")
+    assert not fs.exists(tmpdir_with_protocol + "/afile")
 
-    fs.cp(tmpdir + "/afile3", tmpdir + "/deeply/nested/file")
-    assert fs.exists(tmpdir + "/deeply/nested/file")
+    fs.cp(
+        tmpdir_with_protocol + "/afile3", tmpdir_with_protocol + "/deeply/nested/file"
+    )
+    assert fs.exists(tmpdir_with_protocol + "/deeply/nested/file")
 
-    fs.rm(tmpdir + "/afile3", recursive=True)
-    assert not fs.exists(tmpdir + "/afile3")
+    fs.rm(tmpdir_with_protocol + "/afile3", recursive=True)
+    assert not fs.exists(tmpdir_with_protocol + "/afile3")
 
-    files = [tmpdir + "/afile4", tmpdir + "/afile5"]
+    files = [tmpdir_with_protocol + "/afile4", tmpdir_with_protocol + "/afile5"]
     [fs.touch(f) for f in files]
 
-    with pytest.raises(TypeError):
+    with pytest.raises(AttributeError):
         fs.rm_file(files)
     fs.rm(files)
     assert all(not fs.exists(f) for f in files)
 
-    fs.touch(tmpdir + "/afile6")
-    fs.rm_file(tmpdir + "/afile6")
-    assert not fs.exists(tmpdir + "/afile6")
+    fs.touch(tmpdir_with_protocol + "/afile6")
+    fs.rm_file(tmpdir_with_protocol + "/afile6")
+    assert not fs.exists(tmpdir_with_protocol + "/afile6")
 
     # IsADirectoryError raised on Linux, PermissionError on Windows
     with pytest.raises((IsADirectoryError, PermissionError)):
-        fs.rm_file(tmpdir)
+        fs.rm_file(tmpdir_with_protocol)
 
-    fs.rm(tmpdir, recursive=True)
-    assert not fs.exists(tmpdir)
+    fs.rm(tmpdir_with_protocol, recursive=True)
+    assert not fs.exists(tmpdir_with_protocol)
 
 
 def test_recursive_get_put(tmpdir):
@@ -420,11 +425,11 @@ def test_recursive_get_put(tmpdir):
     fs.touch(tmpdir + "/a1/a2/a3/afile")
     fs.touch(tmpdir + "/a1/afile")
 
-    fs.get("file://{0}/a1".format(tmpdir), tmpdir + "/b1", recursive=True)
+    fs.get(f"file://{tmpdir}/a1", tmpdir + "/b1", recursive=True)
     assert fs.isfile(tmpdir + "/b1/afile")
     assert fs.isfile(tmpdir + "/b1/a2/a3/afile")
 
-    fs.put(tmpdir + "/b1", "file://{0}/c1".format(tmpdir), recursive=True)
+    fs.put(tmpdir + "/b1", f"file://{tmpdir}/c1", recursive=True)
     assert fs.isfile(tmpdir + "/c1/afile")
     assert fs.isfile(tmpdir + "/c1/a2/a3/afile")
 
@@ -465,6 +470,7 @@ def test_make_path_posix():
     )
     if WIN:
         assert make_path_posix("C:\\path") == "C:/path"
+        assert make_path_posix("file://C:\\path\\file") == "C:/path/file"
     if WIN:
         assert (
             make_path_posix(
@@ -480,6 +486,9 @@ def test_make_path_posix():
         )
     assert "/" in make_path_posix("rel\\path")
 
+    pp = make_path_posix("./path")
+    assert "./" not in pp and ".\\" not in pp
+
 
 def test_linked_files(tmpdir):
     tmpdir = str(tmpdir)
@@ -623,6 +632,13 @@ def test_strip_protocol_expanduser():
     assert not LocalFileSystem._strip_protocol("./").endswith("/")
 
 
+def test_strip_protocol_no_authority():
+    path = "file:\\foo\\bar" if WIN else "file:/foo/bar"
+    stripped = LocalFileSystem._strip_protocol(path)
+    assert "file:" not in stripped
+    assert stripped.endswith("/foo/bar")
+
+
 def test_mkdir_twice_faile(tmpdir):
     fn = os.path.join(tmpdir, "test")
     fs = fsspec.filesystem("file")
@@ -636,7 +652,7 @@ def test_iterable(tmpdir):
     fn = os.path.join(tmpdir, "test")
     with open(fn, "wb") as f:
         f.write(data)
-    of = fsspec.open("file://%s" % fn, "rb")
+    of = fsspec.open(f"file://{fn}", "rb")
     with of as f:
         out = list(f)
     assert b"".join(out) == data
@@ -671,7 +687,7 @@ def test_mv_recursive(tmpdir):
 
 @pytest.mark.xfail(WIN, reason="windows expand path to be revisited")
 def test_copy_errors(tmpdir):
-    localfs = fsspec.filesystem("file")
+    localfs = fsspec.filesystem("file", auto_mkdir=True)
 
     dest1 = os.path.join(str(tmpdir), "dest1")
     dest2 = os.path.join(str(tmpdir), "dest2")
@@ -692,6 +708,7 @@ def test_copy_errors(tmpdir):
         localfs.copy([file1, file2, dne], dest1)
 
     localfs.copy([file1, file2, dne], dest1, on_error="ignore")
+
     assert sorted(localfs.ls(dest1)) == [
         make_path_posix(os.path.join(dest1, "afile1")),
         make_path_posix(os.path.join(dest1, "afile2")),
@@ -703,9 +720,9 @@ def test_copy_errors(tmpdir):
     current_files = localfs.expand_path(src, recursive=True)
     with patch.object(localfs, "expand_path", return_value=current_files + [dne]):
         with pytest.raises(FileNotFoundError):
-            localfs.copy(src, dest2, recursive=True, on_error="raise")
+            localfs.copy(src + "/", dest2, recursive=True, on_error="raise")
 
-        localfs.copy(src, dest2, recursive=True)
+        localfs.copy(src + "/", dest2, recursive=True)
         assert sorted(localfs.ls(dest2)) == [
             make_path_posix(os.path.join(dest2, "afile1")),
             make_path_posix(os.path.join(dest2, "afile2")),
@@ -738,6 +755,15 @@ def test_delete_cwd(tmpdir):
         os.chdir(cwd)
 
 
+def test_delete_non_recursive_dir_fails(tmpdir):
+    fs = LocalFileSystem()
+    subdir = os.path.join(tmpdir, "testdir")
+    fs.mkdir(subdir)
+    with pytest.raises(ValueError):
+        fs.rm(subdir)
+    fs.rm(subdir, recursive=True)
+
+
 @pytest.mark.parametrize(
     "opener, ext", [(bz2.open, ".bz2"), (gzip.open, ".gz"), (open, "")]
 )
@@ -785,3 +811,183 @@ def test_numpy_fromfile(tmpdir):
     arr = np.arange(10, dtype=dt)
     arr.tofile(fn)
     assert np.array_equal(np.fromfile(fn, dtype=dt), arr)
+
+
+def test_link(tmpdir):
+    target = os.path.join(tmpdir, "target")
+    link = os.path.join(tmpdir, "link")
+
+    fs = LocalFileSystem()
+    fs.touch(target)
+
+    fs.link(target, link)
+    assert fs.info(link)["nlink"] > 1
+
+
+def test_symlink(tmpdir):
+    target = os.path.join(tmpdir, "target")
+    link = os.path.join(tmpdir, "link")
+
+    fs = LocalFileSystem()
+    fs.touch(target)
+
+    fs.symlink(target, link)
+    assert fs.islink(link)
+
+
+# https://github.com/fsspec/filesystem_spec/issues/967
+def test_put_file_to_dir(tmpdir):
+    src_file = os.path.join(str(tmpdir), "src")
+    target_dir = os.path.join(str(tmpdir), "target")
+    target_file = os.path.join(target_dir, "src")
+
+    fs = LocalFileSystem()
+    fs.touch(src_file)
+    fs.mkdir(target_dir)
+    fs.put(src_file, target_dir)
+
+    assert fs.isfile(target_file)
+
+
+def test_du(tmpdir):
+    file = tmpdir / "file"
+    subdir = tmpdir / "subdir"
+    subfile = subdir / "subfile"
+
+    fs = LocalFileSystem()
+    with open(file, "wb") as f:
+        f.write(b"4444")
+    fs.mkdir(subdir)
+    with open(subfile, "wb") as f:
+        f.write(b"7777777")
+
+    # Switch to posix paths for comparisons
+    tmpdir_posix = Path(tmpdir).as_posix()
+    file_posix = Path(file).as_posix()
+    subdir_posix = Path(subdir).as_posix()
+    subfile_posix = Path(subfile).as_posix()
+
+    assert fs.du(tmpdir) == 11
+    assert fs.du(tmpdir, total=False) == {file_posix: 4, subfile_posix: 7}
+    # Note directory size is OS-specific, but must be >= 0
+    assert fs.du(tmpdir, withdirs=True) >= 11
+
+    d = fs.du(tmpdir, total=False, withdirs=True)
+    assert len(d) == 4
+    assert d[file_posix] == 4
+    assert d[subfile_posix] == 7
+    assert d[tmpdir_posix] >= 0
+    assert d[subdir_posix] >= 0
+
+    assert fs.du(tmpdir, maxdepth=2) == 11
+    assert fs.du(tmpdir, maxdepth=1) == 4
+    with pytest.raises(ValueError):
+        fs.du(tmpdir, maxdepth=0)
+
+    # Size of file only.
+    assert fs.du(file) == 4
+    assert fs.du(file, withdirs=True) == 4
+
+
+@pytest.mark.parametrize("funcname", ["cp", "get", "put"])
+def test_cp_get_put_directory_recursive(tmpdir, funcname):
+    # https://github.com/fsspec/filesystem_spec/issues/1062
+    # Recursive cp/get/put of source directory into non-existent target directory.
+    fs = LocalFileSystem()
+    src = os.path.join(str(tmpdir), "src")
+    fs.mkdir(src)
+    fs.touch(os.path.join(src, "file"))
+
+    target = os.path.join(str(tmpdir), "target")
+
+    if funcname == "cp":
+        func = fs.cp
+    elif funcname == "get":
+        func = fs.get
+    elif funcname == "put":
+        func = fs.put
+
+    # cp/get/put without slash
+    assert not fs.exists(target)
+    for loop in range(2):
+        func(src, target, recursive=True)
+        assert fs.isdir(target)
+
+        if loop == 0:
+            assert fs.find(target) == [make_path_posix(os.path.join(target, "file"))]
+        else:
+            assert sorted(fs.find(target)) == [
+                make_path_posix(os.path.join(target, "file")),
+                make_path_posix(os.path.join(target, "src", "file")),
+            ]
+
+    fs.rm(target, recursive=True)
+
+    # cp/get/put with slash
+    assert not fs.exists(target)
+    for loop in range(2):
+        func(src + "/", target, recursive=True)
+        assert fs.isdir(target)
+        assert fs.find(target) == [make_path_posix(os.path.join(target, "file"))]
+
+
+@pytest.mark.parametrize("funcname", ["cp", "get", "put"])
+def test_cp_get_put_empty_directory(tmpdir, funcname):
+    # https://github.com/fsspec/filesystem_spec/issues/1198
+    # cp/get/put of empty directory.
+    fs = LocalFileSystem(auto_mkdir=True)
+    empty = os.path.join(str(tmpdir), "empty")
+    fs.mkdir(empty)
+
+    target = os.path.join(str(tmpdir), "target")
+    fs.mkdir(target)
+
+    if funcname == "cp":
+        func = fs.cp
+    elif funcname == "get":
+        func = fs.get
+    elif funcname == "put":
+        func = fs.put
+
+    # cp/get/put without slash, target directory exists
+    assert fs.isdir(target)
+    func(empty, target)
+    assert fs.find(target, withdirs=True) == [make_path_posix(target)]
+
+    # cp/get/put with slash, target directory exists
+    assert fs.isdir(target)
+    func(empty + "/", target)
+    assert fs.find(target, withdirs=True) == [make_path_posix(target)]
+
+    fs.rmdir(target)
+
+    # cp/get/put without slash, target directory doesn't exist
+    assert not fs.isdir(target)
+    func(empty, target)
+    assert not fs.isdir(target)
+
+    # cp/get/put with slash, target directory doesn't exist
+    assert not fs.isdir(target)
+    func(empty + "/", target)
+    assert not fs.isdir(target)
+
+
+def test_cp_two_files(tmpdir):
+    fs = LocalFileSystem(auto_mkdir=True)
+    src = os.path.join(str(tmpdir), "src")
+    file0 = os.path.join(src, "file0")
+    file1 = os.path.join(src, "file1")
+    fs.mkdir(src)
+    fs.touch(file0)
+    fs.touch(file1)
+
+    target = os.path.join(str(tmpdir), "target")
+    assert not fs.exists(target)
+
+    fs.cp([file0, file1], target)
+
+    assert fs.isdir(target)
+    assert sorted(fs.find(target)) == [
+        make_path_posix(os.path.join(target, "file0")),
+        make_path_posix(os.path.join(target, "file1")),
+    ]
