@@ -1,15 +1,11 @@
-from __future__ import absolute_import, division, print_function
-
-import datetime
 import zipfile
 
-from fsspec import open_files
+import fsspec
 from fsspec.archive import AbstractArchiveFileSystem
-from fsspec.utils import DEFAULT_BLOCK_SIZE
 
 
 class ZipFileSystem(AbstractArchiveFileSystem):
-    """Read contents of ZIP archive as a file-system
+    """Read/Write contents of ZIP archive as a file-system
 
     Keeps file object open while instance lives.
 
@@ -26,7 +22,9 @@ class ZipFileSystem(AbstractArchiveFileSystem):
         mode="r",
         target_protocol=None,
         target_options=None,
-        block_size=DEFAULT_BLOCK_SIZE,
+        compression=zipfile.ZIP_STORED,
+        allowZip64=True,
+        compresslevel=None,
         **kwargs,
     ):
         """
@@ -34,31 +32,35 @@ class ZipFileSystem(AbstractArchiveFileSystem):
         ----------
         fo: str or file-like
             Contains ZIP, and must exist. If a str, will fetch file using
-            `open_files()`, which must return one file exactly.
+            :meth:`~fsspec.open_files`, which must return one file exactly.
         mode: str
-            Currently, only 'r' accepted
+            Accept: "r", "w", "a"
         target_protocol: str (optional)
             If ``fo`` is a string, this value can be used to override the
             FS protocol inferred from a URL
         target_options: dict (optional)
             Kwargs passed when instantiating the target FS, if ``fo`` is
             a string.
+        compression, allowZip64, compresslevel: passed to ZipFile
+            Only relevant when creating a ZIP
         """
         super().__init__(self, **kwargs)
+        if mode not in set("rwa"):
+            raise ValueError(f"mode '{mode}' no understood")
+        self.mode = mode
         if isinstance(fo, str):
-            files = open_files(
+            fo = fsspec.open(
                 fo, mode=mode + "b", protocol=target_protocol, **(target_options or {})
             )
-            if len(files) != 1:
-                raise ValueError(
-                    'Path "{}" did not resolve to exactly'
-                    'one file: "{}"'.format(fo, files)
-                )
-            fo = files[0]
         self.of = fo
         self.fo = fo.__enter__()  # the whole instance is a context
-        self.zip = zipfile.ZipFile(self.fo, mode=mode)
-        self.block_size = block_size
+        self.zip = zipfile.ZipFile(
+            self.fo,
+            mode=mode,
+            compression=compression,
+            allowZip64=allowZip64,
+            compresslevel=compresslevel,
+        )
         self.dir_cache = None
 
     @classmethod
@@ -67,17 +69,25 @@ class ZipFileSystem(AbstractArchiveFileSystem):
         return super()._strip_protocol(path).lstrip("/")
 
     def __del__(self):
+        if hasattr(self, "zip"):
+            self.close()
+            del self.zip
+
+    def close(self):
+        """Commits any write changes to the file. Done on ``del`` too."""
         self.zip.close()
 
     def _get_dirs(self):
-        if self.dir_cache is None:
+        if self.dir_cache is None or self.mode in set("wa"):
+            # when writing, dir_cache is always in the ZipFile's attributes,
+            # not read from the file.
             files = self.zip.infolist()
             self.dir_cache = {
                 dirname + "/": {"name": dirname + "/", "size": 0, "type": "directory"}
                 for dirname in self._all_dirnames(self.zip.namelist())
             }
             for z in files:
-                f = {s: getattr(z, s) for s in zipfile.ZipInfo.__slots__}
+                f = {s: getattr(z, s, None) for s in zipfile.ZipInfo.__slots__}
                 f.update(
                     {
                         "name": z.filename,
@@ -89,10 +99,7 @@ class ZipFileSystem(AbstractArchiveFileSystem):
 
     def pipe_file(self, path, value, **kwargs):
         # override upstream, because we know the exact file size in this case
-        info = zipfile.ZipInfo(path, datetime.datetime.now().timetuple())
-        info.file_size = len(value)
-        with self.zip.open(path, "w") as f:
-            f.write(value)
+        self.zip.writestr(path, value, **kwargs)
 
     def _open(
         self,
@@ -104,6 +111,12 @@ class ZipFileSystem(AbstractArchiveFileSystem):
         **kwargs,
     ):
         path = self._strip_protocol(path)
+        if "r" in mode and self.mode in set("wa"):
+            if self.exists(path):
+                raise OSError("ZipFS can only be open for reading or writing, not both")
+            raise FileNotFoundError(path)
+        if "r" in self.mode and "w" in mode:
+            raise OSError("ZipFS can only be open for reading or writing, not both")
         out = self.zip.open(path, mode.strip("b"))
         if "r" in mode:
             info = self.info(path)
