@@ -13,6 +13,7 @@ from fsspec.core import (
     _expand_paths,
     expand_paths_if_needed,
     get_compression,
+    get_fs_token_paths,
     open_files,
     open_local,
 )
@@ -75,6 +76,11 @@ def test_expand_error():
         _expand_paths("*.*", None, 1)
 
 
+@pytest.mark.parametrize("mode", ["w", "w+", "x", "x+"])
+def test_expand_fs_token_paths(mode):
+    assert len(get_fs_token_paths("path", mode, num=2, expand=True)[-1]) == 2
+
+
 def test_openfile_api(m):
     m.open("somepath", "wb").write(b"data")
     of = OpenFile(m, "somepath")
@@ -100,7 +106,7 @@ def test_open_local():
     f1 = os.path.join(d1, "f1")
     open(f1, "w").write("test1")
     d2 = str(tempfile.mkdtemp())
-    fn = open_local("simplecache://" + f1, cache_storage=d2, target_protocol="file")
+    fn = open_local(f"simplecache://{f1}", cache_storage=d2, target_protocol="file")
     assert isinstance(fn, str)
     assert open(fn).read() == "test1"
     assert d2 in fn
@@ -144,10 +150,10 @@ def test_pathobject(tmpdir):
 
 def test_automkdir(tmpdir):
     dir = os.path.join(str(tmpdir), "a")
-    of = fsspec.open(os.path.join(dir, "afile"), "w")
-    with of:
-        pass
-    assert "afile" in os.listdir(dir)
+    of = fsspec.open(os.path.join(dir, "afile"), "w", auto_mkdir=False)
+    with pytest.raises(IOError):
+        with of:
+            pass
 
     dir = os.path.join(str(tmpdir), "b")
     of = fsspec.open(os.path.join(dir, "bfile"), "w", auto_mkdir=True)
@@ -300,3 +306,8 @@ def test_chained_url(ftp_writable):
         url += f"::ftp://{username}:{password}@{host}:{port}/archive.zip"
         with fsspec.open(url, "rb") as f:
             assert f.read() == data["afile"]
+
+
+def test_automkdir_local():
+    fs, _ = fsspec.core.url_to_fs("file://", auto_mkdir=True)
+    assert fs.auto_mkdir is True
